@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 
-__version_info__ = (1, 0, 6)
+__version_info__ = (1, 0, 7)
 __version__ = '.'.join(str(i) for i in __version_info__)
 
 import re
@@ -10,17 +10,22 @@ import re
 # See http://www.regular-expressions.info/catastrophic.html
 _url_re = re.compile(r'(?im)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\([^\s()<>]+\))+(?:\([^\s()<>]+\)|[^\s`!()\[\]{};:\'".,<>?]))')
 
+# For the URL tag, try to be smart about when to append a missing http://. If the given link looks like a domain,
+# add a http:// in front of it, otherwise leave it alone (since it may be a relative path, a filename, etc).
+_domain_re = re.compile(r'(?im)(?:www\d{0,3}[.]|[a-z0-9.\-]+[.](?:com|net|org|edu|biz|gov|mil|info|io|name|me|tv|us|uk|mobi))')
+
 class TagOptions (object):
-    tag_name = None                  # The name of the tag, all lowercase.
-    newline_closes = False           # True if a newline should automatically close this tag.
-    standalone = False               # True if this tag does not have a closing tag.
-    render_embedded = True           # True if tags should be rendered inside this tag.
-    transform_newlines = True        # True if newlines should be converted to markup.
-    escape_html = True               # True if HTML characters (<, >, and &) should be escaped inside this tag.
-    replace_links = True             # True if URLs should be replaced with link markup inside this tag.
-    replace_cosmetic = True          # True if cosmetic replacements (elipses, dashes, etc.) should be performed inside this tag.
-    strip = True                     # True if leading and trailing whitespace should be stripped inside this tag.
-    swallow_trailing_newline = False # True if this tag should swallow the first trailing newline (i.e. for block elements).
+    tag_name = None                  #: The name of the tag, all lowercase.
+    newline_closes = False           #: True if a newline should automatically close this tag.
+    same_tag_closes = False          #: True if another start of the same tag should automatically close this tag.
+    standalone = False               #: True if this tag does not have a closing tag.
+    render_embedded = True           #: True if tags should be rendered inside this tag.
+    transform_newlines = True        #: True if newlines should be converted to markup.
+    escape_html = True               #: True if HTML characters (<, >, and &) should be escaped inside this tag.
+    replace_links = True             #: True if URLs should be replaced with link markup inside this tag.
+    replace_cosmetic = True          #: True if cosmetic replacements (elipses, dashes, etc.) should be performed inside this tag.
+    strip = True                     #: True if leading and trailing whitespace should be stripped inside this tag.
+    swallow_trailing_newline = False #: True if this tag should swallow the first trailing newline (i.e. for block elements).
 
     def __init__(self, tag_name, **kwargs):
         self.tag_name = tag_name
@@ -67,8 +72,7 @@ class Parser (object):
         Installs a render function for the specified tag name. The render function
         should have the following signature:
 
-            def render( tag_name, value, options, parent, context ):
-                ...
+            def render(tag_name, value, options, parent, context)
 
         The arguments are as follows:
 
@@ -88,7 +92,7 @@ class Parser (object):
         options = TagOptions(tag_name.strip().lower(), **kwargs)
         self.recognized_tags[options.tag_name] = (render_func, options)
 
-    def add_simple_formatter(self, tag_name, format, **kwargs):
+    def add_simple_formatter(self, tag_name, format_string, **kwargs):
         """
         Installs a formatter that takes the tag options dictionary, puts a value key
         in it, and uses it as a format dictionary to the given format string.
@@ -98,10 +102,15 @@ class Parser (object):
             if options:
                 fmt.update(options)
             fmt.update({'value': value})
-            return format % fmt
+            return format_string % fmt
         self.add_formatter(tag_name, _render, **kwargs)
 
     def install_default_formatters(self):
+        """
+        Installs default formatters for the following tags:
+
+            b, i, u, s, list (and \*), quote, code, center, color, url
+        """
         self.add_simple_formatter('b', '<strong>%(value)s</strong>')
         self.add_simple_formatter('i', '<em>%(value)s</em>')
         self.add_simple_formatter('u', '<u>%(value)s</u>')
@@ -117,7 +126,7 @@ class Parser (object):
             css = ' style="list-style-type:%s;"' % css_opts[list_type] if list_type in css_opts else ''
             return '<%s%s>%s</%s>' % (tag, css, value, tag)
         self.add_formatter('list', _render_list, transform_newlines=False)
-        self.add_simple_formatter('*', '<li>%(value)s</li>', newline_closes=True)
+        self.add_simple_formatter('*', '<li>%(value)s</li>', newline_closes=True, same_tag_closes=True)
         self.add_simple_formatter('quote', '<blockquote>%(value)s</blockquote>')
         self.add_simple_formatter('code', '<code>%(value)s</code>', render_embedded=False)
         self.add_simple_formatter('center', '<div style="text-align:center;">%(value)s</div>')
@@ -136,7 +145,8 @@ class Parser (object):
         self.add_formatter('color', _render_color)
         def _render_url(name, value, options, parent, context):
             href = options['url'] if (options and 'url' in options) else value
-            if '://' not in href:
+            # Only add the missing http:// if it looks like it starts with a domain name.
+            if '://' not in href and _domain_re.match(href):
                 href = 'http://' + href
             return '<a href="%s">%s</a>' % (self._replace(href, self.REPLACE_ESCAPE), value)
         self.add_formatter('url', _render_url, replace_links=False, replace_cosmetic=False)
@@ -259,16 +269,17 @@ class Parser (object):
     def tokenize(self, data):
         """
         Tokenizes the given string. A token is a 4-tuple of the form:
+
             (token_type, tag_name, tag_options, token_text)
 
-        token_type
-            One of: TOKEN_TAG_START, TOKEN_TAG_END, TOKEN_NEWLINE, TOKEN_DATA
-        tag_name
-            The name of the tag if token_type=TOKEN_TAG_*, otherwise None
-        tag_options
-            A dictionary of options specified for TOKEN_TAG_START, otherwise None
-        token_text
-            The original token text
+            token_type
+                One of: TOKEN_TAG_START, TOKEN_TAG_END, TOKEN_NEWLINE, TOKEN_DATA
+            tag_name
+                The name of the tag if token_type=TOKEN_TAG_*, otherwise None
+            tag_options
+                A dictionary of options specified for TOKEN_TAG_START, otherwise None
+            token_text
+                The original token text
         """
         if self.normalize_newlines:
             data = data.replace('\r\n', '\n').replace('\r', '\n')
@@ -315,7 +326,9 @@ class Parser (object):
         Given the current tag options, a list of tokens, and the current position
         in the token list, this function will find the position of the closing token
         associated with the specified tag. This may be a closing tag, a newline, or
-        simply the end of the list (to ensure tags are closed).
+        simply the end of the list (to ensure tags are closed). This function should
+        return a tuple of the form (end_pos, consume), where consume should indicate
+        whether the ending token should be consumed or not.
         """
         embed_count = 0
         while pos < len(tokens):
@@ -323,16 +336,18 @@ class Parser (object):
             if token_type == self.TOKEN_NEWLINE and tag.newline_closes:
                 # If for some crazy reason there are embedded tags that both close on newline,
                 # the first newline will automatically close all those nested tags.
-                return pos
+                return pos, True
             elif token_type == self.TOKEN_TAG_START and tag_name == tag.tag_name:
+                if tag.same_tag_closes:
+                    return pos, False
                 embed_count += 1
             elif token_type == self.TOKEN_TAG_END and tag_name == tag.tag_name:
                 if embed_count > 0:
                     embed_count -= 1
                 else:
-                    return pos
+                    return pos, True
             pos += 1
-        return pos
+        return pos, True
 
     def _link_replace(self, match):
         """
@@ -372,8 +387,11 @@ class Parser (object):
                     formatted.append(render_func(tag_name, None, tag_opts, parent, context))
                 else:
                     # First, find the extent of this tag's tokens.
-                    end = self._find_closing_token(tag, tokens, idx + 1)
+                    end, consume = self._find_closing_token(tag, tokens, idx + 1)
                     subtokens = tokens[idx + 1:end]
+                    # If the end tag should not be consumed, back up one (after grabbing the subtokens).
+                    if not consume:
+                        end = end - 1
                     if tag.render_embedded:
                         # This tag renders embedded tags, simply recurse.
                         inner = self._format_tokens(subtokens, tag, **context)
@@ -389,9 +407,9 @@ class Parser (object):
                     formatted.append(render_func(tag_name, inner, tag_opts, parent, context))
                     # If the tag should swallow the first trailing newline, check the token after the closing token.
                     if tag.swallow_trailing_newline:
-                        next = end + 1
-                        if next < len(tokens) and tokens[next][0] == self.TOKEN_NEWLINE:
-                            end = next
+                        next_pos = end + 1
+                        if next_pos < len(tokens) and tokens[next_pos][0] == self.TOKEN_NEWLINE:
+                            end = next_pos
                     # Skip to the end tag.
                     idx = end
             elif token_type == self.TOKEN_NEWLINE:
@@ -406,10 +424,17 @@ class Parser (object):
         return ''.join(formatted)
 
     def format(self, data, **context):
+        """
+        Formats the input text using any installed renderers. Any context keyword arguments
+        given here will be passed along to the render functions as a context dictionary.
+        """
         tokens = self.tokenize(data)
         return self._format_tokens(tokens, None, **context)
 
     def strip(self, data, strip_newlines=False):
+        """
+        Strips out any tags from the input text, using the same tokenization as the formatter.
+        """
         text = []
         for token_type, tag_name, tag_opts, token_text in self.tokenize(data):
             if token_type == self.TOKEN_DATA:
@@ -420,7 +445,7 @@ class Parser (object):
 
 g_parser = None
 
-def render_html(input, **context):
+def render_html(input_text, **context):
     """
     A module-level convenience method that creates a default bbcode parser,
     and renders the input string as HTML.
@@ -428,7 +453,7 @@ def render_html(input, **context):
     global g_parser
     if g_parser is None:
         g_parser = Parser()
-    return g_parser.format(input, **context)
+    return g_parser.format(input_text, **context)
 
 if __name__ == '__main__':
     import sys
